MODULE CodeGen;		(*27.01.2000 LB, based on NW, 17.11.94*)

	FROM VirtRisc IMPORT
		WordSize, MemSize, OpCode, Memory, FP, SP, RT, RegSet,
		opL,opPos, aL,aPos,bL,bPos,cPos,cL1;
	FROM Scanner IMPORT 
		Symbol;
	IMPORT Word, Scanner, SIO, VirtRisc;
	
	CONST
		MaxRel	=	NUMBER(Memory) DIV 4;
		RIDiff	=	ORD(OpCode.ADDI) - ORD(OpCode.ADD);

	TYPE
		Rel	=	ARRAY [0 .. MaxRel - 1] OF INTEGER;

		ErrKind	=	{badType, valueLarge, badLevel, 
				 badIndexType,badIndexVal,
				 illegalAssignment, incompAssignment, 
				 illegalParMode, badParType, 
				 isBoolean, isInteger, noRegs};
	CONST
		ErrText	=	ARRAY ErrKind OF TEXT {
				"bad type", "value too large", "bad level", 
				"bad indext type (not an integer)", "bad index value",
				"illegal assignment", "incompatible assignment",
				"illegal parameter mode", "bad parameter type", 
				"Boolean?", "Integer?", "Not enough registers, sorry!"};
		ResRegs	=	RegSet{0, FP, SP, RT};	(*0 contains always 0*)
				(*FP: frame pointer, SP: stack pointer, rt: return adr*)


	VAR
		pc:			INTEGER;	(*program counter*)
		curlev, relx:		INTEGER;
		regs:			RegSet;		(*actually used regsiters*)
		code:			Memory;
		rel:			Rel;

	PROCEDURE PC(): INTEGER = BEGIN RETURN pc END PC;

	PROCEDURE CurrentLevel(): INTEGER = BEGIN RETURN curlev END CurrentLevel;

	PROCEDURE Mark(msg: TEXT) =
	BEGIN
		Scanner.Mark("CodeGenerator => " & msg);
	END Mark;	

	PROCEDURE GetReg(VAR r: INTEGER) =
	VAR i := 1;	(*Register 0 is reserved for 0*)
	BEGIN
		WHILE (i < FP) AND (i IN regs) DO INC(i) END;
		IF i IN regs THEN Mark(ErrText[ErrKind.noRegs]) END;
		regs:= regs + RegSet{i};
		r:= i;
	END GetReg;


	PROCEDURE Put(op: OpCode; a, b, c: INTEGER) =
	(*Emit instruction*)
	VAR cw: Word.T;			(*codeWord*)
	BEGIN
		cw:= Word.Insert(cw, ORD(op), opPos, opL);
		cw:= Word.Insert(cw, a, aPos, aL);
		cw:= Word.Insert(cw, b, bPos, bL);
		cw:= Word.Insert(cw, c, cPos, cL1);
		code[pc]:= cw;
		INC(pc);
	END Put;

	PROCEDURE TestRange(x: INTEGER) =
	(*Check 16 bit entity*)
	BEGIN
		IF (x >= 16_8000) OR (x < -16_8000) THEN Mark(ErrText[ErrKind.valueLarge]) END;
	END TestRange;

	PROCEDURE Load(VAR x: Item) =
	VAR r: INTEGER;
	BEGIN
		<*ASSERT x.mode # Mode.Reg*>
		IF x.mode = Mode.Var THEN
			IF x.lev = 0 THEN rel[relx]:= pc; INC(relx) END;
			GetReg(r);
			Put(OpCode.LDW, r, x.r, x.a);
			regs:= regs - RegSet{x.r};
			x.r:= r
		ELSIF x.mode = Mode.Const THEN
			IF x.a = 0 THEN
				x.r:= 0
			ELSE
				TestRange(x.a);
				GetReg(x.r);
				Put(OpCode.ADDI, x.r, 0, x.a)
			END
		END;
		x.mode:= Mode.Reg;
	END Load;

	PROCEDURE LoadBool(VAR x: Item) =
	BEGIN
		<*ASSERT x.mode # Mode.Jmp*>
		IF x.mode = Mode.Var THEN
			Load(x);
			Put(OpCode.BEQ, x.r, 0, 0);
		ELSIF x.mode = Mode.Const THEN
			IF x.a = 0 THEN Put(OpCode.BEQ, 0, 0, 0)ELSE Put(OpCode.BNE, 0, 0, 0) END;
		END;
		x.mode:= Mode.Jmp;
		x.a:= pc - 1;
		x.b:= 0
	END LoadBool;

	PROCEDURE PutOp(cd: OpCode; VAR x, y: Item) =
	VAR r: INTEGER;
	BEGIN
		IF x.mode # Mode.Reg THEN Load(x) END;
		IF x.r = 0 THEN GetReg(x.r); r:= 0 ELSE r:= x.r	END;
		IF y.mode = Mode.Const THEN 
			TestRange(y.a); 
			INC(cd, RIDiff);	(*change opcode to opcodeI*)
			Put(cd, r, x.r, y.a)
		ELSE
			IF y.mode # Mode.Reg THEN Load(y) END;
			Put(cd, x.r, r, y.r);
			regs:= regs - RegSet{y.r}
		END;
	END PutOp;

	PROCEDURE Negate(op: [OpCode.BEQ .. OpCode.BGT]): OpCode =
	BEGIN
		CASE	op	OF
		| OpCode.BEQ	=> RETURN OpCode.BNE;
		| OpCode.BNE	=> RETURN OpCode.BEQ;
		| OpCode.BLT	=> RETURN OpCode.BGE;
		| OpCode.BGE	=> RETURN OpCode.BLT;
		| OpCode.BLE	=> RETURN OpCode.BGT;
		| OpCode.BGT	=> RETURN OpCode.BLE;
		END;
	END Negate;	

	PROCEDURE NegPostCond(VAR x: Item) =
	VAR L, cd: INTEGER; op: OpCode;
	BEGIN
		L:= x.a;
		cd:= code[L];
		(*x.a:= cd MOD HalfW;*)
		x.a:= Word.Extract(cd, cPos, cL1);			(*Extracts low 16 bit*)
		cd:= cd - x.a + x.b;
		x.b:= L;
		op:= VAL(Word.Extract(cd, opPos, opL), OpCode);		(*Extracts op code*)	
		cd:= Word.Insert(cd, ORD(Negate(op)), opPos, opL);	(*Rewrites negated op code*)
		code[L]:= cd
	END NegPostCond;

	PROCEDURE Merged(L0, L1: INTEGER): INTEGER =
	VAR L2, L3: INTEGER;
	BEGIN
		IF L0 # 0 THEN
			L2:= L0;
			LOOP
				(*L3:= code[L2] MOD HalfW;*)
				L3:= Word.Extract(code[L2], cPos, cL1); (*Extracts low 16 bit*)
				IF L3 = 0 THEN EXIT END;
				L2:= L3
			END; (*LOOP*)
			code[L2]:= code[L2] - L3 + L1;
			RETURN L0
		ELSE
			RETURN L1
		END (*IF L0 # 0*)
	END Merged;

	PROCEDURE Fix(at, with: INTEGER) =
	BEGIN
		(*code[at]:= code[at] DIV HalfW * HalfW + (with MOD HalfW)*)
		code[at]:= Word.Insert(code[at], with, cPos, cL1);  (*"with" replaces low 16 bits*)
	END Fix;

	PROCEDURE CondFixup(L: INTEGER) =
	BEGIN
		IF L = pc - 1 THEN DEC(pc) ELSE Fix(L, pc - L) END
	END CondFixup;

	PROCEDURE FixWith(L0, L1: INTEGER) =
	VAR L2: INTEGER;
	BEGIN
		WHILE L0 # 0 DO 
			L2:= Word.Extract(code[L0], cPos, cL1);
			(*L2:= code[L0] MOD HalfW;*)
			Fix(L0, L1 - L0);
			L0:= L2
		END (*WHILE L0 # 0*)
	END FixWith;

	PROCEDURE FixLink(L: INTEGER) =
	VAR L1: INTEGER;
	BEGIN
		WHILE L # 0 DO 
			L1:= Word.Extract(code[L], cPos, cL1);
			(*L1:= code[L] MOD HalfW;*)
			Fix(L, pc - L);
			L:= L1;
		END (*WHILE L # 0*)
	END FixLink;

	(*---------------------------------------------------------*)

	PROCEDURE IncLevel(n: INTEGER := 1) =
	BEGIN
		INC(curlev, n)
	END IncLevel;

	PROCEDURE MakeConstItem(VAR x: Item; type: Type; val: INTEGER) =
	BEGIN
		x.mode:= Mode.Const; x.type:= type; x.a:= val
	END MakeConstItem;

	PROCEDURE MakeItem(VAR x: Item; y: Object) =
	VAR r: INTEGER;
	BEGIN
		x.mode:= y.class; x.type:= y.type; x.lev:= y.lev; x.a:= y.val;
		IF y.lev = 0 THEN		x.r:= 0
		ELSIF y.lev = curlev THEN	x.r:= FP
		ELSE				Mark(ErrText[ErrKind.badLevel]); x.r:= 0
		END;
		IF y.class = Mode.Par THEN
			GetReg(r);
			Put(OpCode.LDW, r, x.r, x.a);
			x.mode:= Mode.Var; x.r:= r; x.a:= 0
		END;
	END MakeItem;


	PROCEDURE Field(VAR x: Item; y: Object) =
	(*x:= x.y*)
	BEGIN
		INC(x.a, y.val); x.type:= y.type
	END Field;	


	PROCEDURE Index(VAR x, y: Item) =
	(*x:= x[y]*)
	BEGIN
		IF y.type # intType THEN Mark(ErrText[ErrKind.badIndexType]) END;
		IF y.mode = Mode.Const THEN
			IF (y.a < 0) OR (y.a >= x.type.len) THEN Mark(ErrText[ErrKind.badIndexVal]) END;
			INC(x.a, y.a * x.type.base.size)
		ELSE
			IF y.mode = Mode.Var THEN Load(y) END;
			Put(OpCode.MULI, y.r, y.r, x.type.base.size);
			IF x.r # 0 THEN 
				Put(OpCode.ADD, y.r, x.r, y.r);
				regs:= regs - RegSet{x.r};
			END;
			x.r:= y.r
		END;
		x.type:= x.type.base
	END Index;

	PROCEDURE Op1(op: Symbol; VAR x: Item) =
	(*x:= op x*)
	VAR t: INTEGER;
	BEGIN
		IF op = Symbol.minus THEN
			IF x.type.form # Form.Integer THEN Mark(ErrText[ErrKind.badType])
			ELSIF x.mode = Mode.Const THEN
				x.a:= - x.a
			ELSE
				IF x.mode = Mode.Var THEN Load(x) END;
				Put(OpCode.SUB, x.r, 0, x.r);
			END
		ELSIF op = Symbol.not THEN
			IF x.mode # Mode.Jmp THEN LoadBool(x) END;
			NegPostCond(x);
			t:= x.a; x.a:= x.b; x.b:= t
		ELSIF op = Symbol.and THEN
			IF x.mode # Mode.Jmp THEN LoadBool(x) END;
			FixLink(x.b);
			x.b:= 0;
		ELSIF op = Symbol.or THEN
			IF x.mode # Mode.Jmp THEN LoadBool(x) END;
			NegPostCond(x);
			FixLink(x.a);
			x.a:= 0;
		END; (*IF op*)
	END Op1;

	PROCEDURE Op2(op: Symbol; VAR x, y: Item) =
	(*x:= x op y*)
	BEGIN
		IF (x.type.form = Form.Integer) AND (y.type.form = Form.Integer) THEN
			IF (x.mode = Mode.Const) AND (y.mode = Mode.Const) THEN
				CASE op OF
				| Symbol.plus	=>	INC(x.a, y.a);
				| Symbol.minus	=>	DEC(x.a, y.a);
				| Symbol.times	=>	x.a:= x.a * y.a;
				| Symbol.div	=>	x.a:= x.a DIV y.a;
				| Symbol.mod	=>	x.a:= x.a MOD y.a;
				ELSE			Mark(ErrText[ErrKind.badType])
				END; (*CASE op*)
			ELSE
				CASE op OF
				| Symbol.plus	=>	PutOp(OpCode.ADD, x, y);
				| Symbol.minus	=>	PutOp(OpCode.SUB, x, y);	
				| Symbol.times	=>	PutOp(OpCode.MUL, x, y);	
				| Symbol.div	=>	PutOp(OpCode.Div, x, y);	
				| Symbol.mod	=>	PutOp(OpCode.Mod, x, y);	
				ELSE			Mark(ErrText[ErrKind.badType])
				END; (*CASE op*)
			END (*IF x.mode*)
		ELSIF (x.type.form = Form.Boolean) AND (y.type.form = Form.Boolean) THEN
			IF y.mode # Mode.Jmp THEN LoadBool(y) END;
			IF op = Symbol.or THEN
				x.a:= y.a;
				x.b:= Merged(y.b, x.b)
			ELSIF op = Symbol.and THEN
				x.b:= y.b;
				x.a:= Merged(y.a, x.a)
			END (*IF op*)
		ELSE
			Mark(ErrText[ErrKind.badType])
		END; (*IF x.type...*)
	END Op2;

	PROCEDURE Relation(op: [Symbol.eql .. Symbol.geq]; VAR x, y: Item) =
	(*x:= x rel y*)
	BEGIN
		IF (x.type.form # Form.Integer) OR (y.type.form # Form.Integer) THEN
			Mark(ErrText[ErrKind.badType])
		ELSE
			PutOp(OpCode.CMP, x, y);
			IF x.mode # Mode.Reg THEN Load(x) END;
			IF y.mode # Mode.Reg THEN Load(y) END;
			CASE op OF
			| Symbol.eql	=>	Put(OpCode.BNE, x.r, y.r, 0);
			| Symbol.neq	=>	Put(OpCode.BEQ, x.r, y.r, 0);
			| Symbol.lss	=>	Put(OpCode.BGE, x.r, y.r, 0);
			| Symbol.leq	=>	Put(OpCode.BGT, x.r, y.r, 0);
			| Symbol.gtr	=>	Put(OpCode.BLE, x.r, y.r, 0);
			| Symbol.geq	=>	Put(OpCode.BLT, x.r, y.r, 0);
			END; (*CASE op*)
			regs:= regs - RegSet{x.r};
			regs:= regs - RegSet{y.r};
		END; (*IF x.type...*)
		x.mode:= Mode.Jmp; x.type:= boolType; x.a:= pc - 1; x.b:= 0
	END Relation;

	PROCEDURE Store(VAR x, y: Item) =
	CONST IntBool = SET OF Form{Form.Integer, Form.Boolean};
	BEGIN
		IF (x.type.form IN IntBool) AND (x.type.form = y.type.form) THEN
			IF y.mode = Mode.Jmp THEN
				FixLink(y.b);
				GetReg(y.r);
				Put(OpCode.ADDI, y.r, 0, 1);
				Put(OpCode.BEQ, 0, 0, 2);
				FixLink(y.a);
				Put(OpCode.ADDI, y.r, 0, 0);
			ELSIF	y.mode # Mode.Reg THEN 
				Load(y);
			END; (*IF y.mode = Mode.Jmp*)
			IF x.mode = Mode.Var THEN
				IF x.lev = 0 THEN rel[relx]:= pc; INC(relx) END;
				Put(OpCode.STW, y.r, x.r, x.a);
			ELSE
				Mark(ErrText[ErrKind.illegalAssignment]);
			END; (*IF y.mode...*)
			regs:= regs - RegSet{x.r};
			regs:= regs - RegSet{y.r};
		ELSE
			Mark(ErrText[ErrKind.incompAssignment]);
		END			
	END Store;	

	PROCEDURE Parameter(VAR x: Item; ftype: Type; class: Mode) =
	VAR r: INTEGER;
	BEGIN
		IF x.type = ftype THEN
			IF class = Mode.Par THEN		(*Var parameter*)
				IF x.mode = Mode.Var THEN
					IF x.a # 0 THEN
						IF x.lev = 0 THEN rel[relx]:= pc; INC(relx) END;
						GetReg(r);
						Put(OpCode.ADDI, r, x.r, x.a);
					ELSE
						r:= x.r
					END (*IF x.a...*)
				ELSE
					Mark(ErrText[ErrKind.illegalParMode])
				END; (*IF x.mode...*)
				Put(OpCode.PUSH, r, SP, WordSize);
				regs:= regs - RegSet{r}
			ELSE						(*Value parameter*)
				IF x.mode # Mode.Reg THEN Load(x) END;
				Put(OpCode.PUSH, x.r, SP, WordSize);
				regs:= regs - RegSet{x.r}
			END; (*IF class...*)
		ELSE
			Mark(ErrText[ErrKind.badParType])
		END; (*IF x.type...*)
						
	END Parameter;

	(*---------------------------------------------------------*)


	PROCEDURE TestBool(VAR x: Item) =
	BEGIN
		IF x.type.form = Form.Boolean THEN
			IF x.mode = Mode.Jmp THEN FixLink(x.b) ELSE LoadBool(x) END
		ELSE
			Mark(ErrText[ErrKind.isBoolean]);
			x.a:= pc;
			Put(OpCode.BEQ, 0, 0, 0)
		END
	END TestBool;


	PROCEDURE TestBoolR(VAR x: Item; l: INTEGER) =
	BEGIN
		IF x.type.form = Form.Boolean THEN
			FixWith(x.a, l);
			FixLink(x.b)
		ELSE
			Mark(ErrText[ErrKind.isBoolean])
		END
	END TestBoolR;

	PROCEDURE BJump(L: INTEGER) =
	BEGIN
		Put(OpCode.BEQ, 0, 0, L - pc);
	END BJump;

	PROCEDURE FJump(VAR L: INTEGER) =
	BEGIN
		Put(OpCode.BEQ, 0, 0, L);
		L:= pc - 1;
	END FJump;

	PROCEDURE Call(VAR x: Item) =
	BEGIN
		Put(OpCode.BSR, 0, 0, x.a - pc);
	END Call;

	PROCEDURE IOCall(VAR x, y: Item) =
	VAR z: Item;
	BEGIN
		IF x.a < 4 THEN
			IF y.type.form # Form.Integer THEN Mark(ErrText[ErrKind.isInteger]) END
		END;
		IF x.a = 1 THEN
			GetReg(z.r);
			z.mode:= Mode.Reg; z.type:= intType;
			Put(OpCode.RD, z.r, 0, 0);
			Store(y, z);
		ELSIF x.a = 2 THEN
			Load(y);
			Put(OpCode.WRD, 0, 0, y.r);
			regs:= regs - RegSet{y.r}
		ELSIF x.a = 3 THEN
			Load(y);
			Put(OpCode.WRH, 0, 0, y.r);
			regs:= regs - RegSet{y.r}
		ELSE
			Load(y);
			Put(OpCode.WRC, 0, 0, y.r);
			regs:= regs - RegSet{y.r}
		END (*IF x.a...*)
	END IOCall;

	PROCEDURE Enter(size: INTEGER) =
	BEGIN
		IF curlev > 0 THEN
			Put(OpCode.PUSH, RT, SP, WordSize);	(*Push retrun adr*)
			Put(OpCode.PUSH, FP, SP, WordSize);	(*Push FP*)
			Put(OpCode.ADD, FP, 0, SP);		(*FP:= SP*)
			Put(OpCode.SUBI, SP, SP, size);		(*SP:= SP - size*)
		ELSE
			Put(OpCode.ADDI, SP, 0, MemSize-size);	(*SP:= Mem size - size*)
		END (*IF curlev*)
	END Enter;

	PROCEDURE Return(size: INTEGER) =
	BEGIN
		Put(OpCode.ADD, SP, 0, FP);			(*SP:= FP*)
		Put(OpCode.POP, FP, SP, WordSize);		(*POP FP*)
		Put(OpCode.POP, RT, SP, WordSize + size);	(*POP retrun adr*)
		Put(OpCode.RET, 0, 0, RT);			(*return*)
	END Return;

	PROCEDURE Open() =
	BEGIN
		curlev:= 0; pc:= 0; relx:= 0; regs:= RegSet{};
	END Open;

	PROCEDURE Relocate() =
	(*CONST H = 16_10000;*)
	VAR k, c: INTEGER;
	BEGIN
		FOR i:= 0 TO relx - 1 DO
			k:= rel[i];
			c:= Word.Extract(code[k], 0, 16);
			code[k]:= Word.Insert(code[k], c + MemSize, 0, 16);
			(*code[k]:= (code[k] DIV H * H) +
					(code[k] + MemSize) MOD H;*)
		END; (*FOR i*)
	END Relocate;

	PROCEDURE Close() =
	BEGIN
		Put(OpCode.EXT, 0, 0, 0);
	END Close;

	PROCEDURE LoadAndGo(trace, stepwise: BOOLEAN := TRUE; wr: SIO.Writer := NIL) =
	BEGIN
		Relocate();
		VirtRisc.Execute(code, trace, stepwise, pc, 0, wr);
	END LoadAndGo;

BEGIN
	boolType:= NEW(Type, form:= Form.Boolean, size:= WordSize);
	intType:= NEW(Type, form:= Form.Integer, size:= WordSize);
	regs:= ResRegs;
END CodeGen.