MODULE Parser;	(*27.01.2000 LB, based on NW, 17.11.94*)

	  IMPORT Text, SIO, SF, Scanner, CodeGen;
	  
	  FROM Scanner 
	  IMPORT Symbol, GetSym, GetId, GetValue, Error, SetRd, SetErr;
	  
	  FROM CodeGen 
	  IMPORT Object, Item, Mode, Form, Type, boolType, intType;

	  FROM VirtRisc
	  IMPORT WordSize;

	  TYPE
	  		 ErrKind	=	{multiDef, undef, noArray, norBrak,
	  		 			    noRec, noId, norParen, nolParen, noStatment, noBecomes,
	  		 			    nolpcom, noFactor, manyPars, fewPars,
	  		 			    illAssignment, noThen, noEnd, noUntil, noDo, noSemi,
	  		 			    noType, badIndex, noOf, noDecl, noEql, noColon,
	  		 			    noConst, badDecSeq,
	  		 			    structPar, noMatch, noDot, noMod};	
	  CONST
	  		 ErrText	=	ARRAY ErrKind OF TEXT 
	  		 			  {"multiple definition", "undefined", "not an array", "] ?",
							"not a record", "ident ?", ") ?", "( ?", "statement ?", ":= ?",
			 			  	") or , ?", "factor? ", "too many parameters", "too few parameters",
							"illegal assignment ?", "THEN ?", "END ?", "UNTIL ?", "DO ? ", "; ?",
							"type ?", "bad index", "OF ?", "declaration ?", "= ?",  ": ?",
							"expression not a constant", "bad declaration sequence",
							"no structured parameters", "name does not match", ". ?", "MODULE ?"};

	  VAR
	  		 sym:	Symbol;
	  		 topScope, universe, guard: Object := NIL;


	  PROCEDURE Mark(msg: TEXT) =
	  BEGIN
	  		 Scanner.Mark("Parser => " & msg);
	  END Mark;


	  PROCEDURE Check(expected: Symbol; errKind: ErrKind) =
	  BEGIN
	  		 IF sym = expected THEN 
	  		 		sym:= GetSym() 
	  		 ELSE
	  		 		Mark(ErrText[errKind])
	  		 END; (*IF sym = expected *)
	  END Check;

	  PROCEDURE Check2(expected: SET OF Symbol; errKind: ErrKind) =
	  BEGIN
	  		 IF NOT (sym IN expected) THEN
	  		 		Mark(ErrText[errKind]);
	  		 		REPEAT sym:= GetSym(); UNTIL sym IN expected;
	  		 END (*IF NOT (sym IN expected)*)
	  END Check2;


	  PROCEDURE NewObj(VAR obj: Object; cl: Mode) =
	  VAR id: TEXT := GetId(); new, x: Object;
	  BEGIN
	  		 x:= topScope;
	  		 guard.name:= id;
	  		 WHILE NOT Text.Equal(x.next.name, id) DO x:= x.next END;
	  		 		IF x.next = guard THEN
	  		 			  new:= NEW(Object, name:= id, class:= cl, next:= guard);
	  		 			  x.next:= new; obj:= new
	  		 		ELSE
	  		 			  obj:= x.next;
	  		 			  Mark(ErrText[ErrKind.multiDef]);
	  		 		END; (*IF*)
	  END NewObj;	
   


	  PROCEDURE Find(VAR obj: Object) =
	  VAR id: TEXT := GetId(); s, x: Object;
	  BEGIN
	  		 s:= topScope;
	  		 guard.name:= id;
	  		 LOOP
	  		 		x:= s.next;
	  		 		WHILE NOT Text.Equal(x.name, id) DO 
	  		 			  x:= x.next;
	  		 		END;
	  		 		IF x # guard THEN obj:= x; EXIT END;
	  		 		IF s = universe THEN
	  		 			  obj:= x; Mark(ErrText[ErrKind.undef]); 
	  		 			  EXIT
	  		 		END; (*IF s*)
	  		 		s:= s.dsc;
	  		 END; (*LOOP*)
	  END Find;



	  PROCEDURE FindField(VAR obj: Object; list: Object) =
	  VAR id: TEXT := GetId();
	  BEGIN
	  		 guard.name:= id;
	  		 WHILE NOT Text.Equal(list.name, id) DO list:= list.next END;
	  		 		obj:= list;
	  		 END FindField;



	  PROCEDURE IsParam(obj: Object): BOOLEAN =
	  BEGIN
			 	  RETURN (obj.class = Mode.Par) OR (obj.class = Mode.Var) AND (obj.val > 0)
	  END IsParam;



	  PROCEDURE OpenScope() =
	  VAR s := NEW(Object, class:= Mode.Head, dsc:= topScope, next:= guard);
	  BEGIN
	  		 topScope:= s
	  END OpenScope;

	  

	  PROCEDURE CloseScope() =
	  BEGIN
	  		 topScope:= topScope.dsc
	  END CloseScope;



	  PROCEDURE Selector(VAR x: Item) =
	  CONST Sels = SET OF Symbol{Symbol.lbrak, Symbol.period};
	  VAR y: Item; obj: Object;
	  BEGIN
	  		 WHILE sym IN Sels DO
	  		 		IF sym = Symbol.lbrak THEN	(*[ arrived*)
						  sym:= GetSym();
	  		 			  Expression(y);
				IF x.type.form = Form.Array THEN
					CodeGen.Index(x, y)
				ELSE
					Mark(ErrText[ErrKind.noArray]);
				END; (*IF x...*)
				Check(Symbol.rbrak,  ErrKind.norBrak);
			ELSE				(*. arrived*)
				sym:= GetSym();
				IF sym = Symbol.ident THEN
					IF x.type.form = Form.Record THEN
						FindField(obj, x.type.fields);
						sym:= GetSym();
						IF obj # guard THEN
							CodeGen.Field(x, obj)
						ELSE
							Mark(ErrText[ErrKind.undef])
						END (*IF obj # guared*)
					ELSE
						Mark(ErrText[ErrKind.noRec])
					END (*IF x.type...*)
				ELSE
					Mark(ErrText[ErrKind.noId])
				END (*IF sym = Symbol.ident*)
			END(*IF sym = Symbol.lbrak*)
		END (*WHILE sym IN Sels*)
	END Selector;


	PROCEDURE Factor(VAR x: Item) =
	CONST Expected = SET OF Symbol{Symbol.lparen .. Symbol.eof};
	VAR obj: Object;
	BEGIN (*sync*)
		Check2(Expected, ErrKind.noId);
		IF sym = Symbol.ident THEN
			Find(obj);
			sym:= GetSym();
			CodeGen.MakeItem(x, obj);
			Selector(x)
		ELSIF sym = Symbol.number THEN
			CodeGen.MakeConstItem(x, intType, GetValue());
			sym:= GetSym();
		ELSIF sym = Symbol.lparen THEN
			sym:= GetSym();
			Expression(x);
			Check(Symbol.rparen, ErrKind.norParen);
		ELSIF sym = Symbol.not THEN
			sym:= GetSym();
			Factor(x);			
			CodeGen.Op1(Symbol.not, x)
		ELSE
			Mark(ErrText[ErrKind.noFactor]);
			CodeGen.MakeItem(x, guard);
		END (*IF sym...*)
	END Factor;

	PROCEDURE Term(VAR x: Item) =
	CONST Muls = SET OF Symbol{Symbol.times .. Symbol.and};
	VAR y: Item; op: Symbol;
	BEGIN
		Factor(x);
		WHILE sym IN Muls DO
			op:= sym;
			sym:= GetSym();
			IF op = Symbol.and THEN CodeGen.Op1(op, x) END;
			Factor(y);
			CodeGen.Op2(op, x, y)
		END (*WHILE sym IN Muls*)
	END Term;

	PROCEDURE SimpleExpression(VAR x: Item) =
	CONST Adds = SET OF Symbol{Symbol.plus .. Symbol.or};
	VAR y: Item; op: Symbol;
	BEGIN
		IF sym = Symbol.plus THEN
			sym:= GetSym();
			Term(x)		
		ELSIF sym = Symbol.minus THEN
			sym:= GetSym();
			Term(x);
			CodeGen.Op1(Symbol.minus, x)
		ELSE
			Term(x)		
		END; (*IF sym...*)
		WHILE sym IN Adds DO
			op:= sym;
			sym:= GetSym();
			IF op = Symbol.or THEN CodeGen.Op1(op, x) END;
			Term(y);
			CodeGen.Op2(op, x, y)	
		END (*WHILE sym IN Adds*)
	END SimpleExpression;

	PROCEDURE Expression(VAR x: Item) =
	CONST Rels = SET OF Symbol{Symbol.eql .. Symbol.geq};
	VAR y: Item; op: Symbol; 
	BEGIN
		SimpleExpression(x);
		IF sym IN Rels THEN
			op:= sym;
			sym:= GetSym();
			SimpleExpression(y);
			CodeGen.Relation(op, x, y)	
		END; (*IF sym...*)
	END Expression;

	PROCEDURE Parameter(VAR fp: Object) =
	VAR x: Item;
	BEGIN
		Expression(x);
		IF IsParam(fp) THEN
			CodeGen.Parameter(x, fp.type, fp.class);
			fp:= fp.next
		ELSE
			Mark(ErrText[ErrKind.manyPars])
		END (*IF IsParam(fp)*)
	END Parameter;

	PROCEDURE StatSequence() =
	CONST Expected = SET OF Symbol{Symbol.ident .. Symbol.eof};
	VAR par, obj: Object; x, y: Item; L: INTEGER;

		PROCEDURE Param(VAR x: Item) =
		BEGIN
			Check(Symbol.lparen , ErrKind.nolParen);
			Expression(x);
			Check(Symbol.rparen , ErrKind.norParen);
		END Param;

	BEGIN (*StatSequence*)
		LOOP (*sync*)
			obj:= guard;
			Check2(Expected, ErrKind.noStatment);
			CASE sym OF
			| Symbol.ident	=>
				Find(obj);
				sym:= GetSym();
				CodeGen.MakeItem(x, obj);
				Selector(x);
				IF sym = Symbol.becomes THEN
					sym:= GetSym();
					Expression(y);
					CodeGen.Store(x, y);
				ELSIF sym = Symbol.eql THEN
					Mark(ErrText[ErrKind.noBecomes]);
					sym:= GetSym();
					Expression(y);
				ELSIF x.mode = Mode.Proc THEN
					par:= obj.dsc;
					IF sym = Symbol.lparen THEN 
						sym:= GetSym();
						IF sym = Symbol.rparen THEN
							sym:= GetSym()
						ELSE
							LOOP
								Parameter(par);
								IF sym = Symbol.comma THEN
									sym:= GetSym()
								ELSIF sym = Symbol.rparen THEN
									sym:= GetSym();
									EXIT
								ELSIF sym >= Symbol.semicolon THEN
									EXIT
								ELSE
									Mark(ErrText[ErrKind.nolpcom]);
								END (*IF sym = Symbol.comma*)
							END (*LOOP*)
						END; (*IF sym = Symbol.rparen*)
					END; (*IF sym = Symbol.lparen*)
					IF NOT IsParam(par) THEN
						CodeGen.Call(x)
					ELSE
						Mark(ErrText[ErrKind.fewPars])
					END (* NOT IsParam(par) *)
				ELSIF x.mode = Mode.SProc THEN
					IF obj.val <= 4 THEN Param(y) END;
					CodeGen.IOCall(x, y)
				ELSIF obj.class = Mode.Typ THEN
					Mark(ErrText[ErrKind.illAssignment])	
				END; (*IF sym = Symbol.becomes*)
			| Symbol.if	=>
				sym:= GetSym();
				Expression(x);
				CodeGen.TestBool(x);
				Check(Symbol.then , ErrKind.noThen);
				StatSequence();
				L:= 0;
				WHILE sym = Symbol.elsif  DO
					sym:= GetSym();
					CodeGen.FJump(L);
					CodeGen.FixLink(x.a);
					Expression(x);
					CodeGen.TestBool(x);
					Check(Symbol.then , ErrKind.noThen);
					StatSequence()
				END; (*WHILE sym = Symbol.elsif*)
				IF sym = Symbol.else THEN
					sym:= GetSym();
					CodeGen.FJump(L);
					CodeGen.FixLink(x.a);
					StatSequence()
				ELSE
					CodeGen.FixLink(x.a);
				END; (*IF sym = Symbol.else*)
				CodeGen.FixLink(L);
				Check(Symbol.end , ErrKind.noEnd);
			| Symbol.repeat	=>
				sym:= GetSym();
				L:= CodeGen.PC();
				StatSequence();
				Check(Symbol.until , ErrKind.noUntil);
				Expression(x);
				CodeGen.TestBoolR(x, L);
			| Symbol.while	=>
				sym:= GetSym();
				L:= CodeGen.PC();
				Expression(x);
				CodeGen.TestBool(x);
				Check(Symbol.do , ErrKind.noDo);
				StatSequence();
				CodeGen.BJump(L);
				CodeGen.FixLink(x.a);
				Check(Symbol.end , ErrKind.noEnd);
			ELSE	
			END; (*CASE sym OF*)
			IF sym = Symbol.semicolon THEN
				sym:= GetSym()
			ELSIF sym IN SET OF Symbol{Symbol.end .. Symbol.elsif, Symbol.array .. 
Symbol.eof, Symbol.until} THEN
				EXIT
			ELSE
				Mark(ErrText[ErrKind.noSemi])
			END (*sym = Symbol.semicolon*)
		END (*LOOP*)
	END StatSequence;

	PROCEDURE IdentList(class: Mode; VAR first: Object) =
	VAR obj: Object;
	BEGIN
		IF sym = Symbol.ident THEN
			NewObj(first, class);
			sym:= GetSym();
			WHILE sym = Symbol.comma DO
				sym:= GetSym();
				IF sym = Symbol.ident THEN
					NewObj(obj, class);
					sym:= GetSym()
				ELSE
					Mark(ErrText[ErrKind.noId])
				END (*IF sym = Symbol.ident*)
			END; (*WHILE sym = Symbol.comma*)
			Check(Symbol.colon, ErrKind.noColon);
		END (*IF sym = Symbol.ident*)
	END IdentList;

	PROCEDURE ProcType(VAR type: Type) =
	CONST Expected = SET OF Symbol{
		Symbol.ident, Symbol.array .. Symbol.eof};
	VAR obj, first: Object; x: Item; tp: Type;
	BEGIN
		type:= intType;
		Check2(Expected, ErrKind.noType);
		CASE sym OF
		| Symbol.ident	=>
			Find(obj);
			sym:= GetSym();
			IF obj.class = Mode.Typ THEN 
				type:= obj.type;
			ELSE 
				Mark(ErrText[ErrKind.noType]) 
			END;
		| Symbol.array	=>
			sym:= GetSym();
			Expression(x);
			IF (x.mode # Mode.Const) OR (x.a < 0) THEN Mark(ErrText[ErrKind.badIndex]) 
END;
			Check(Symbol.of, ErrKind.noOf);
			ProcType(tp);
			type:= NEW(Type, form:= Form.Array, base:= tp, len:= x.a, size:= x.a * 
tp.size);
		| Symbol.record	=>
			sym:= GetSym();
			type:= NEW(Type, form:= Form.Record, size:= 0);
			OpenScope();
			LOOP
				IF sym = Symbol.ident THEN
					IdentList(Mode.Fld, first);
					ProcType(tp);
					obj:= first;
					WHILE obj # guard DO
						obj.type:= tp; obj.val:= type.size;
						INC(type.size, obj.type.size);
						obj:= obj.next
					END; (*WHILE obj # guard*)
				END;(*IF sym = Symbol.ident*)
				IF	sym = Symbol. semicolon THEN sym:= GetSym()
				ELSIF	sym = Symbol.ident THEN Mark(ErrText[ErrKind.noSemi])
				ELSE	EXIT
				END;(*IF sym = Symbol.semicolon *)
			END; (*LOOP*)
			type.fields:= topScope.next;
			CloseScope();
			Check(Symbol.end, ErrKind.noEnd);
		ELSE
			Mark(ErrText[ErrKind.noId])
		END; (*CASE sym OF*)
	END ProcType;

	PROCEDURE Declarations(varsize: INTEGER) =
        	CONST 
		Expected	= SET OF Symbol{Symbol.end, Symbol.const .. Symbol.eof};
		DecSyms	= SET OF Symbol{Symbol.const, Symbol.type, Symbol.var};
	VAR 
		obj, first: Object;
		x: Item; tp: Type; L: INTEGER;
	BEGIN
		Check2(Expected, ErrKind.noDecl);
		LOOP
			IF sym = Symbol.const THEN
				sym:= GetSym();
				WHILE sym = Symbol.ident DO
					NewObj(obj, Mode.Const);
					sym:= GetSym();
					Check(Symbol.eql, ErrKind.noEql);
					Expression(x);
					IF x.mode = Mode.Const THEN
                			                                obj.val:= x.a; obj.type:= 
x.type
					ELSE
						Mark(ErrText[ErrKind.noConst])
					END; (*x.mode = Mode.Const*)
					Check(Symbol.semicolon, ErrKind.noSemi);
				END; (*WHILE sym = Symbol.ident*)
			END; (*sym = Symbol.const*)
			IF sym = Symbol.type THEN
				sym:= GetSym();
				WHILE sym = Symbol.ident DO
					NewObj(obj, Mode.Typ);
					sym:= GetSym();
					Check(Symbol.eql, ErrKind.noEql);
					ProcType(obj.type);
					Check(Symbol.semicolon, ErrKind.noSemi);
				END; (*WHILE sym = Symbol.ident*)
			END; (*sym = Symbol.type*)
			IF sym = Symbol.var THEN
				sym:= GetSym();
				WHILE sym = Symbol.ident DO
					IdentList(Mode.Var, first);
					ProcType(tp);
 					obj:= first;
					WHILE obj # guard DO
    						obj.type:= tp; 
						obj.lev:= CodeGen.CurrentLevel();
 						INC(varsize, obj.type.size);	
						obj.val:= -varsize;	 obj:= obj.next;
					END; (*WHILE obj # gurad*)
					Check(Symbol.semicolon, ErrKind.noSemi);
				END; (*WHILE sym = Symbol.ident*)
			END; (*sym = Symbol.var*)	
			IF sym IN DecSyms THEN
				Mark(ErrText[ErrKind.badDecSeq])	
			ELSE
				EXIT
			END; (*IF sym IN...*)
  		END; (*LOOP*)
		CodeGen.Enter(varsize);
		L:= 0;
		CodeGen.FJump(L);
		WHILE sym = Symbol.procedure DO
			ProcedureDecl();
			Check(Symbol.semicolon, ErrKind.noSemi);
		END; (*sym = symbol.procedure*)
		CodeGen.CondFixup(L);
	END Declarations;

	PROCEDURE ProcedureDecl() =
	CONST marksize = 2 * WordSize;
	VAR
		proc, obj: Object;
		procid: TEXT;
		locblksize, parblksize: INTEGER;

		PROCEDURE FPSection() =
		CONST StructForm = SET OF Form{Form.Array, Form.Record};
		VAR obj, first: Object; tp: Type; parsize: INTEGER;
		BEGIN
			IF sym = Symbol.var THEN		(*VAR Parameter*)
				sym:= GetSym();
				IdentList(Mode.Par, first)
			ELSE
				IdentList(Mode.Var, first)		(*VALUE Parameter*)
			END; (*IF sym = Symbol.var*)
			IF sym = Symbol.ident THEN
				Find(obj);
				sym:= GetSym();
				IF obj.class = Mode.Typ THEN
					tp:= obj.type
				ELSE
					Mark(ErrText[ErrKind.noType]);
					tp:= intType;
				END; (*sym = Symbol.ident*)
			ELSE
				Mark(ErrText[ErrKind.noId]);
				tp:= intType;
			END; (*sym = Symbol.ident*)
			IF first.class = Mode.Var THEN		(*VALUE Parameter*)
				parsize:= tp.size;
				IF tp.form IN StructForm THEN Mark(ErrText[ErrKind.structPar]) END;
			ELSE					(*VAR Parameter*)
				parsize:= WordSize
			END; (*first.class = Mode.Var*)
			obj:= first;
			WHILE obj # guard DO
				obj.type:= tp; INC(parblksize, parsize); obj:= obj.next
			END (*WHILE obj # guard*)
		END FPSection;

	BEGIN (*ProcedureDecl*)
		sym:= GetSym();
		IF sym = Symbol.ident THEN
			procid:= GetId();
			NewObj(proc, Mode.Proc);
			sym:= GetSym();
			parblksize:= marksize;
			CodeGen.IncLevel(1);
			OpenScope();
			proc.val:= CodeGen.PC();
			IF sym = Symbol.lparen THEN
				sym:= GetSym();
				IF sym = Symbol.rparen THEN
					sym:= GetSym()
				ELSE
					FPSection();
					WHILE sym = Symbol.semicolon DO 
						sym:= GetSym(); 
						FPSection();
					END; (*WHILE sym = Symbol.semicolon*)
					Check(Symbol.rparen, ErrKind.norParen);
				END; (*IF sym = Symbol.rparen*)
			END; (*IF sym = Symbol.lparen*)
			obj:= topScope.next;
			locblksize:= parblksize;
			WHILE obj # guard DO
				obj.lev:= CodeGen.CurrentLevel();
				IF obj.class = Mode.Par THEN DEC(locblksize, WordSize)
				ELSE DEC(locblksize, obj.type.size)
				END; (*IF obj.class = Mode.Par *)
				obj.val:= locblksize;
				obj:= obj.next
			END; (*WHILE obj # guard*)
			proc.dsc:= topScope.next;
			Check(Symbol.semicolon, ErrKind.noSemi);
			locblksize:= 0;
			Declarations(locblksize);
			IF sym  = Symbol.begin THEN sym:= GetSym(); StatSequence() END;
			Check(Symbol.end, ErrKind.noEnd);
			IF sym = Symbol.ident THEN
				IF NOT Text.Equal(procid, GetId()) THEN 
					Mark(ErrText[ErrKind.noMatch]) 
				END; (*IF NOT Text...*)
				sym:= GetSym();
			END; (*IF sym = Symbol.ident*)
			CodeGen.Return(parblksize - marksize);
			CloseScope();
			CodeGen.IncLevel(-1)
		END; (*IF sym = Symbol.ident*)
	END ProcedureDecl;

	PROCEDURE Module() =
	VAR modid: TEXT; varsize: INTEGER;
	BEGIN
		IF sym = Symbol.module THEN
			sym:= GetSym();
			CodeGen.Open();
			OpenScope();
			varsize:= 0;
			IF sym = Symbol.ident THEN
				modid:= GetId();
				sym:= GetSym();
			ELSE
				Mark(ErrText[ErrKind.noId])
			END; (*IF sym = Symbol.ident*)
			Check(Symbol.semicolon, ErrKind.noSemi);
			Declarations(varsize);
			IF sym = Symbol.begin THEN 
				sym:= GetSym(); 
				StatSequence() ;
			END;
			IF sym = Symbol.end THEN 
				sym:= GetSym()
			ELSE
				Mark(ErrText[ErrKind.noEnd])
			END; (*sym = Symbol.end*)
			IF sym = Symbol.ident THEN
				IF NOT Text.Equal(modid, GetId()) THEN 
					Mark(ErrText[ErrKind.noMatch]) 
				END; (*IF NOT Text...*)
				sym:= GetSym();
			ELSE
				Mark(ErrText[ErrKind.noId]) 
			END; (*IF sym = Symbol.ident*)
			IF sym # Symbol.period THEN Mark(ErrText[ErrKind.noDot]) END;
			CloseScope();
			CodeGen.Close();
		ELSE
			Mark(ErrText[ErrKind.noMod]) 
		END; (*IF sym = Symbol.module*)			
	END Module;

	PROCEDURE Enter(cl: Mode; n: INTEGER; nm: TEXT; tp: Type) =
	VAR obj:= NEW(Object, class:= cl, val:= n, name:= nm, type:= tp, dsc:= NIL, 
next:= topScope.next);
	BEGIN
		topScope.next:= obj
	END Enter;

	PROCEDURE Init()=
	BEGIN
		guard:= NEW(Object, class:= Mode.Var, type:= intType, val:= 0);
		OpenScope();
		Enter(Mode.Typ, 1, "BOOLEAN", boolType);
		Enter(Mode.Typ, 2, "INTEGER", intType);
		Enter(Mode.Const, 0, "FALSE", boolType);
		Enter(Mode.Const, 1, "TRUE", boolType);
		Enter(Mode.SProc, 1, "Read", NIL);
		Enter(Mode.SProc, 2, "Write", NIL);
		Enter(Mode.SProc, 3, "WriteHex", NIL);
		Enter(Mode.SProc, 4, "WriteChar", NIL);
		universe:= topScope
	END Init;

	PROCEDURE Start(inFile: SIO.Reader; errFile, asmFile: SIO.Writer := NIL;
		trace, stepwise: BOOLEAN:= TRUE) =
	BEGIN
		Init();
		SetRd(inFile);
		SetErr(errFile);
		sym:= GetSym();
		Module();
		IF NOT Error() THEN
			CodeGen.LoadAndGo(trace, stepwise, asmFile)
		ELSE
			SF.CloseWrite(errFile)
		END; (*IF NOT Error()*)
	END Start;

BEGIN
END Parser.
