MODULE Scanner;  				(*25.01.2000 LB based on NW 17.11.94*)

	  IMPORT SIO, Text, Fmt;

	  TYPE
	  		 KeyRec	=	RECORD sym: Symbol; kw: TEXT END;
	   CONST	KeyTab	=	ARRAY OF KeyRec 
				{KeyRec{sym:= Symbol.do, kw:= "DO"}, 
				 KeyRec{sym:= Symbol.until, kw:= "UNTIL"}, 
				 KeyRec{sym:= Symbol.if, kw:= "IF"}, 
				 KeyRec{sym:= Symbol.of, kw:= "OF"}, 
				 KeyRec{sym:= Symbol.or, kw:= "OR"}, 
				 KeyRec{sym:= Symbol.end, kw:= "END"}, 
				 KeyRec{sym:= Symbol.mod, kw:= "MOD"}, 
				 KeyRec{sym:= Symbol.var, kw:= "VAR"}, 
				 KeyRec{sym:= Symbol.else, kw:= "ELSE"}, 
				 KeyRec{sym:= Symbol.then, kw:= "THEN"}, 
				 KeyRec{sym:= Symbol.type, kw:= "TYPE"}, 
				 KeyRec{sym:= Symbol.array, kw:= "ARRAY"}, 
				 KeyRec{sym:= Symbol.begin, kw:= "BEGIN"}, 
				 KeyRec{sym:= Symbol.const, kw:= "CONST"}, 
				 KeyRec{sym:= Symbol.elsif, kw:= "ELSIF"}, 
				 KeyRec{sym:= Symbol.repeat, kw:= "REPEAT"}, 
				 KeyRec{sym:= Symbol.while, kw:= "WHILE"}, 
				 KeyRec{sym:= Symbol.record, kw:= "RECORD"}, 
				 KeyRec{sym:= Symbol.procedure, kw:= "PROCEDURE"}, 
				 KeyRec{sym:= Symbol.div, kw:= "DIV"}, 
				 KeyRec{sym:= Symbol.module, kw:= "MODULE"}  };
		Digits	=	SET OF CHAR {'0' .. '9'};
		Letters	=	SET OF CHAR {'A' .. 'Z', 'a' .. 'z'};
		Singles	=	SET OF CHAR {'&', '*', '+', '-', '=', '#',
				';', ',', '.', ')', '[', ']', '~'};	

	VAR
		rd:	SIO.Reader := NIL;		(*Source file*)
		wr:	SIO.Writer := NIL;		(*Error file*)
		ch:	CHAR    	:= ' ';		(*Actual character*)
		val:	INTEGER;			(*Actual value of number*)
		id:	TEXT;				(*Actual identifier*)
		error:	BOOLEAN  	:= FALSE;
		line, linePos: CARDINAL := 0; (*line and row counter*)

	PROCEDURE GetValue():	INTEGER = 
	BEGIN 
		RETURN val 
	END GetValue;
	PROCEDURE GetId(): TEXT = 
	BEGIN 
		RETURN id 
	END GetId;
	PROCEDURE Error(): BOOLEAN = 
	BEGIN 
		RETURN error 
	END Error;
	PROCEDURE SetRd(inFile: SIO.Reader)= 
	BEGIN 
		rd:= inFile 
	END SetRd;
	PROCEDURE SetErr(errFile: SIO.Writer) = 
	BEGIN 
		wr:= errFile 
	END SetErr;

	PROCEDURE GetChar(rd: SIO.Reader): CHAR =
	CONST lf = '\n'; ret = '\r';
	VAR ch: CHAR;
	BEGIN
		TRY
			ch:= SIO.GetChar(rd);
			IF ch = ret THEN ch:= SIO.GetChar(rd) END;
			INC(linePos);
			IF ch = lf THEN INC(line); linePos:= 0 END;
		EXCEPT
			SIO.Error =>
				Mark("Undefined error on the input file")
		END; (*TRY*)
		RETURN ch
	END GetChar;

	PROCEDURE Mark(msg: TEXT) =
	BEGIN
		SIO.PutLine("[" & Fmt.Int(line+1) & ", " &
			Fmt.Int(linePos+1)  & "] : " & msg, wr);
		error:= TRUE;
	END Mark;

	PROCEDURE GetSym(): Symbol =
	VAR sym:	Symbol;

		PROCEDURE Ident() =
		BEGIN
			id:= "";
			REPEAT
				id:= id & Text.FromChar(ch);
				ch:= GetChar(rd);	(*forward input*)
			UNTIL NOT (SIO.End(rd) OR (ch IN Letters + Digits));
			sym:= Symbol.ident;	(*suppose it is a user-identifier*)
			FOR i:= FIRST(KeyTab) TO LAST(KeyTab) DO
				IF Text.Equal(KeyTab[i].kw, id) THEN
					sym:= KeyTab[i].sym (*it is a keyword*)
				END; (*IF*)
			END (*FOR i*)
		END Ident;

		PROCEDURE Number() =
		BEGIN
			val:= 0;
			sym:= Symbol.number;
			REPEAT
				IF val <= (LAST(INTEGER) - ORD(ch) 
						+ ORD('0')) DIV 10 THEN
					val:= 10 * val + (ORD(ch) - ORD('0'))
				ELSE
					Mark("number too large");
					val:= 0;
				END; (*IF val*)
				ch:= GetChar(rd);	(*forward input*)
			UNTIL NOT (ch IN Digits);
		END Number;

		PROCEDURE Comment() =
		BEGIN
			ch:= GetChar(rd);	(*forward input, consume '*'*)
			LOOP
				LOOP
					WHILE ch = '(' DO
						ch:= GetChar(rd);	(*forward input*)
						IF ch = '*' THEN Comment() END
					END; (*WHILE*)
					IF ch = '*' THEN ch:= GetChar(rd); EXIT END;
					IF SIO.End(rd) THEN EXIT END;
					ch:= GetChar(rd);	(*forward input*)
				END; (*LOOP*)
				IF ch = ')' THEN ch:= GetChar(rd); EXIT END;
				IF SIO.End(rd) THEN 
					Mark("Comment not terminated"); EXIT; 
				END
			END; (*LOOP*)			
		END Comment;

	BEGIN (*GetSym*)
		WHILE (NOT SIO.End(rd)) AND (ch <= ' ') DO  (*Skip blanks*)
			ch:= GetChar(rd)
		END; (*WHILE*)
		IF SIO.End(rd) THEN sym:= Symbol.eof; RETURN sym END;
		IF ch IN Singles THEN
			CASE	ch	OF
			| '&'	=>	sym:= Symbol.and;
			| '*'	=>	sym:= Symbol.times; 
			| '+'	=>	sym:= Symbol.plus;
			| '-'	=>	sym:= Symbol.minus;
			| '='	=>	sym:= Symbol.eql;
			| '#'	=>	sym:= Symbol.neq;
			| ';'	=>	sym:= Symbol.semicolon;
			| ','	=>	sym:= Symbol.comma;
			| '.'	=>	sym:= Symbol.period;
			| ')'	=>	sym:= Symbol.rparen;
			| '['	=>	sym:= Symbol.lbrak;
			| ']'	=>	sym:= Symbol.rbrak;
			| '~'	=>	sym:= Symbol.not;
			END; (*CASE ch*)
			ch:=  GetChar(rd);			(*forward input*)
		ELSIF ch IN Digits THEN
			Number();
		ELSIF ch IN Letters THEN
			Ident();
		ELSE
			CASE	ch	OF
			| '<'	=>
				ch:=  GetChar(rd);		(*forward input*)
				IF ch = '=' THEN
					sym:= Symbol.leq;
					ch:=  GetChar(rd);	(*forward input*)
				ELSE
					sym:= Symbol.lss;
				END; (*IF ch*)
			| '>'	=>
				ch:=  GetChar(rd);		(*forward input*)
				IF ch = '=' THEN
					sym:= Symbol.geq;
					ch:=  GetChar(rd);	(*forward input*)
				ELSE
					sym:= Symbol.gtr;
				END; (*IF ch*)

			| ':'	=>
				ch:=  GetChar(rd);		(*forward input*)
				IF ch = '=' THEN
					sym:= Symbol.becomes;
					ch:=  GetChar(rd);	(*forward input*)
				ELSE
					sym:= Symbol.colon;
				END; (*IF ch*)
			| '('	=>
				ch:=  GetChar(rd);		(*forward input*)
				IF ch = '*' THEN
					Comment();
					sym:= GetSym();
				ELSE
					sym:= Symbol.lparen;
				END; (*IF ch*)
			ELSE
				sym:= Symbol.null;
				ch:=  GetChar(rd);		(*forward input*)
			END; (*CASE ch*)
		END; (*IF ch IN ..*)
		RETURN sym;
	END GetSym;

BEGIN
END Scanner.
