MODULE VirtRisc;	(*27.01.2000 LB, based on NW, 17.11.94*)


	IMPORT Word, SIO, Text, SF;
	FROM SIO IMPORT
		PutInt, Nl, PutText, PutChar, Writer;

	CONST
		CReg	= Vars + OpSet{OpCode.CHK};
		F3 	= OpSet {OpCode.JSR};
		Noa	= Calls + Writes + OpSet{OpCode.EXT};
		Nob	= Noa + Checks + Read;
		Noc	= Read;

	VAR					(*Virtual special registers*)
		PC:	INTEGER;		(*Program Counter*)
		IR:	INTEGER;		(*Instruction Register*)
		R:	ARRAY Regs OF INTEGER;	(*Register Set*)

	PROCEDURE Trap(msg: TEXT) =
	BEGIN
		SIO.PutLine(msg);
	END Trap;

	PROCEDURE DisplayInst(op: OpCode; a, b, c: INTEGER; at: INTEGER; 
			      wr: Writer; running: BOOLEAN := FALSE) =
	CONST tab = "\t"; NumL = 6;
	VAR empty: BOOLEAN;

		PROCEDURE PutReg(r: INTEGER; content, empty: BOOLEAN) =
		CONST	Specials	= RegSet{FP, SP, RT};
			SpecNames	= ARRAY [FP .. RT] OF TEXT {
					"FP", "SP", "RT"};
		BEGIN
			IF empty THEN
				FOR i:= 1 TO 2 * NumL + 2 DO PutChar(' ') END
			ELSIF r IN Specials THEN
				PutText("    " & SpecNames[r], wr)
			ELSE
				PutInt(r, NumL, wr);
			END; (*IF r IN Specials*)
			IF (NOT empty) AND content THEN
				PutText("[");
				PutInt(R[r], NumL, wr);
				PutText("]");
			END;
			PutText(tab, wr) 
		END PutReg;	

	BEGIN (*DisplayInst*)			
		Nl(wr);
		PutInt( 4 * at, NumL, wr);
		PutText(tab & OpName[op] & tab, wr);
		empty:= op IN Noa;
		PutReg(a, running, empty);
		empty:= op IN (Noa + Nob);		
		PutReg(b, running, empty);
		empty:= op IN Noc;		
		IF running AND op IN (Branch + Calls - F3) THEN 
			c:= (c + PC) * WordSize
		END;
		PutReg(c, FALSE, empty);
	END DisplayInst;

	PROCEDURE DecodeInst(ir: Word.T; VAR opCode: OpCode; VAR a, b, c: INTEGER) =
	BEGIN
		opCode:= VAL(Word.Extract(ir, opPos, opL), OpCode);	
							(*Most significant 6 bits*)
		a:=  Word.Extract(ir, aPos, aL);	(*Next 5 bits*)
		b:= Word.Extract(ir, bPos, bL);		(*Next 5 bits*)
		IF opCode IN F3 THEN
			c:= Word.Extract(ir, cPos, cL2)
		ELSE	
			c:= Word.Extract(ir, cPos, cL1);(*Least significant 16 bits*)
			IF c >= b16 THEN DEC(c, signE)END;	(*sign extention*)
		END; (*IF opCode IF F3*)
	END DecodeInst;

	PROCEDURE Decode(READONLY M: Memory; pc, start: INTEGER; wr: SIO.Writer) =
	VAR  opCode: OpCode; a, b, c: INTEGER;
	BEGIN
		FOR i:= start TO pc - 1 DO
			DecodeInst(M[i], opCode, a, b, c);
			DisplayInst(opCode, a, b, c, i, wr, running:= FALSE);
		END; (*FOR*)
		SF.CloseWrite(wr);
	END Decode;


	PROCEDURE Execute(VAR M: Memory; trace, stepwise: BOOLEAN:= FALSE;
			  pc: INTEGER; start: INTEGER := 0;  wr: SIO.Writer := NIL) =
	CONST 
	VAR
		opCode: OpCode;
		a, b, c, next: INTEGER;
		line: TEXT;
	BEGIN
		Decode(M, pc, start, wr);
		R[RT]:= 0;
		PC:= start DIV WordSize;
		LOOP
			R[0]:= 0;
			next:= PC + 1;
			IR:= M[PC];
			DecodeInst(IR, opCode, a, b, c);
			IF opCode IN CReg THEN c:= R[c MOD NumRegs] END; (*c:= R.c*)
			CASE	opCode	OF
			(*Register Operations*)
			| OpCode.ADD, OpCode.ADDI	=>	R[a]:= R[b] + c;
			| OpCode.SUB, OpCode.SUBI,
			  OpCode.CMP, OpCode.CMPI	=>	R[a]:= R[b] - c;
			| OpCode.MUL, OpCode.MULI	=>	R[a]:= R[b] * c;
			| OpCode.Div, OpCode.DIVI	=>	R[a]:= R[b] DIV c;
			| OpCode.Mod, OpCode.MODI	=>	R[a]:= R[b] MOD c;
			| OpCode.Or,  OpCode.ORI	=>	R[a]:= Word.Or(R[b], c);
			| OpCode.And, OpCode.ANDI	=>	R[a]:= Word.And(R[b], c);
			| OpCode.BIC, OpCode.BICI	=>	R[a]:= Word.And(R[b], Word.Not(c));
			| OpCode.XOR, OpCode.XORI	=>	R[a]:= Word.Xor(R[b], c);
			| OpCode.SHL, OpCode.SHLI	=>	R[a]:= Word.Shift(R[b], c);
			| OpCode.SHA, OpCode.SHAI	=>	R[a]:= Word.Shift(R[b], c);
				IF c < 0 THEN			(*arithmetic shift to the right*)
					R[a]:= R[b]; FOR i:= 1 TO ABS(c) DO R[a] := R[a] DIV 2 END;
				END; (*IF c < 0*)
			| OpCode.CHK, OpCode.CHKI	=>	
				IF (R[a] < 0) OR (R[a] >= c) THEN Trap("CHK error") END;
			(*Load and Store instructions*)
			| OpCode.LDW	=>	R[a]:= M[(R[b] + c) DIV WordSize];
			| OpCode.LDB	=>	Trap("LDB not implemented"); EXIT;
			| OpCode.POP	=>	R[a]:= M[R[b] DIV WordSize]; INC(R[b], c);
			| OpCode.STW	=>	M[(R[b] + c) DIV WordSize]:= R[a];
			| OpCode.STB	=>	Trap("STB not implemented"); EXIT;
			| OpCode.PUSH	=>	DEC(R[b], c); M[R[b] DIV WordSize]:= R[a]
			(*Conditional, PC-relative jumps*)
			| OpCode.BEQ	=>	IF R[a] = 0 THEN next:= PC + c END;
			| OpCode.BNE	=>	IF R[a] # 0 THEN next:= PC + c END;
			| OpCode.BLT	=>	IF R[a] < 0 THEN next:= PC + c END;
			| OpCode.BLE	=>	IF R[a] <= 0 THEN next:= PC + c END;
			| OpCode.BGT	=>	IF R[a] > 0 THEN next:= PC + c END;
			| OpCode.BGE	=>	IF R[a] >= 0 THEN next:= PC + c END;
			(*Subroutine call (PC-relative and absolute) and return*)
			| OpCode.BSR	=>	next:= PC + c; R[RT]:= (PC + 1) * WordSize;	
			| OpCode.JSI	=>	next:= c; R[RT]:= (PC + 1) * WordSize;	
			| OpCode.JSR	=>	next:= c; R[RT]:= (PC + 1) * WordSize;	
			| OpCode.RET	=>	next:= R[c MOD NumRegs] DIV WordSize;
						IF next = 0 THEN EXIT END;
			(*Special I/O instructions - for test purposes*)	
			| OpCode.RD	=>	R[a]:= SIO.GetInt();	
			| OpCode.WRD	=>	SIO.PutInt(R[c], 6);	
			| OpCode.WRH	=>	SIO.PutUnsigned(R[c], 6, 16);	
			| OpCode.WRC	=>	SIO.PutChar(VAL(R[c], CHAR));	
			(*EXIT*)	
			| OpCode.EXT	=>	EXIT;	
			END; (*CASE opCode*)
			IF trace THEN DisplayInst(opCode, a, b, c, PC, NIL, running:= TRUE) END;
			IF stepwise THEN 
				line:= SIO.GetLine();
				IF Text.FindChar(line, 'r') >= 0 THEN stepwise:= FALSE END
			END; (*IF stepwise*)
			PC:= next;
		END; (*LOOP*)
		SIO.PutLine("\nVirtual Risc ends, good by");
	END Execute;
    	

BEGIN
END VirtRisc.